""" Register Controller Definition """
from typing import Optional

from starlette import status
from starlette.exceptions import HTTPException

from rowantree.auth.sdk import RegisterUserRequest, User, UserBase

from .abstract_controller import AbstractController


class RegisterController(AbstractController):
    """Register Controller"""

    def execute(self, request: RegisterUserRequest) -> UserBase:
        """
        Create new user.

        Parameters
        ----------
        request: RegisterUserRequest
            Form data for new user creation.
        """

        user: Optional[User] = self.auth_service.register_user(request=request)

        if not user:
            raise HTTPException(status_code=status.HTTP_409_CONFLICT, detail="Unable to create user")

        return UserBase.parse_obj(user)



""" Abstract Controller Definition """

from rowantree.contracts import BaseModel

from .db.dao import DBDAO


class AbstractService(BaseModel):
    """
    Abstract Service

    Attributes
    ----------
    dao: DBDAO
        The database DAO.
    """

    dao: DBDAO



""" Token Controller Definition """

from typing import Optional

from fastapi.security import OAuth2PasswordRequestForm
from starlette import status
from starlette.exceptions import HTTPException

from rowantree.auth.sdk import Token, User, create_user_access_token

from .abstract_controller import AbstractController


class TokenController(AbstractController):
    """Token Controller"""

    def execute(self, request: OAuth2PasswordRequestForm) -> Token:
        """
        Execute the controller action. (Auth User)

        Parameters
        ----------
        request: OAuth2PasswordRequestForm
            The username/password form data

        Returns
        -------
        token: Token
            A token with the users permissions.
        """

        user: Optional[User] = self.auth_service.authenticate_user(request.username, request.password)

        if not user:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Incorrect username or password",
                headers={"WWW-Authenticate": "Bearer"},
            )
        return create_user_access_token(user=user)



""" Database DAO Definition """

import logging
import socket
from typing import Optional, Tuple

import mysql.connector
from mysql.connector import errorcode
from mysql.connector.pooling import MySQLConnectionPool, PooledMySQLConnection
from starlette import status
from starlette.exceptions import HTTPException

from rowantree.auth.sdk import User
from rowantree.contracts import BaseModel

from ...contracts.duplicate_key_error import DuplicateKeyError
from .incorrect_row_count_error import IncorrectRowCountError


class DBDAO(BaseModel):
    """
    Database DAO

    Attributes
    ----------
    cnxpool: Any
        MySQL Connection Pool
    """

    cnxpool: MySQLConnectionPool

    # pylint: disable=duplicate-code
    def _call_proc(self, name: str, args: list, debug: bool = False) -> Optional[list[Tuple]]:
        """
        Perform a stored procedure call.

        Parameters
        ----------
        name: str
            The name of the stored procedure to call.
        args: list
            The arguments to pass to the stored procedure.
        debug: bool
            Whether to log debug details about the call.

        Returns
        -------
        results: Optional[list[Tuple]]
            An optional list of tuples (rows) from the call.
        """

        if debug:
            logging.debug("[DAO] [Stored Proc Call Details] Name: {%s}, Arguments: {%s}", name, args)
        rows: Optional[list[Tuple]] = None
        cnx: Optional[PooledMySQLConnection] = None
        try:
            cnx = self.cnxpool.get_connection()
            cursor = cnx.cursor()
            cursor.callproc(name, args)
            for result in cursor.stored_results():
                rows = result.fetchall()
            cursor.close()
            cnx.close()
        except socket.error as error:
            logging.error("socket.error")
            logging.error(error)
            if cnx:
                cnx.close()
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Internal Server Error"
            ) from error
        except mysql.connector.Error as error:
            logging.error("mysql.connector.Error: {%i}", error.errno)
            logging.error(str(error))
            if cnx:
                cnx.close()
            if error.errno == errorcode.ER_ACCESS_DENIED_ERROR:
                logging.error("Something is wrong with your user name or password")
            elif error.errno == errorcode.ER_BAD_DB_ERROR:
                logging.error("Database does not exist")
            elif error.errno == errorcode.ER_DUP_ENTRY:
                logging.error("Duplicate key")
                raise DuplicateKeyError(str(error)) from error
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Internal Server Error"
            ) from error
        except Exception as error:
            logging.error("error")
            # All other uncaught exception types
            logging.error(str(error))
            if cnx:
                cnx.close()
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Internal Server Error"
            ) from error

        if debug:
            logging.debug("[DAO] [Stored Proc Call Details] Returning:")
            logging.debug(rows)
        return rows

    def get_user(self, username: Optional[str] = None, guid: Optional[str] = None) -> User:
        """
        Get User (From Database) Using `username` first, else `guid`.  At least one must be supplied.

        Parameters
        ----------
        username: Optional[str]
            username
        guid: Optional[str]
            user guid

        Returns
        -------
        user: User
            User (from database)
        """

        if username is None and guid is None:
            logging.debug("get_user called without username or a guid, one is required")
            raise HTTPException(status_code=status.HTTP_500_INTERNAL_SERVER_ERROR)

        args: list[str] = [username]
        proc_name: str = "getUserByUsername"

        if username is None:
            args = [guid]
            proc_name: str = "getUserByGUID"

        rows: Optional[list[Tuple]] = self._call_proc(proc_name, args, True)
        if rows is None or len(rows) != 1:
            raise IncorrectRowCountError(f"Result count was not exactly one. Received: {rows}")
        user: tuple = rows[0]

        is_disabled: bool = True
        if user[6] == 0:
            is_disabled = False

        is_admin: bool = False
        if user[7] == 1:
            is_admin = True

        return User(
            username=user[2], guid=user[1], email=user[3], hashed_password=user[4], disabled=is_disabled, admin=is_admin
        )

    def create_user(self, user: User) -> User:
        """
        Creates a new user within the database.

        Parameters
        ----------
        user: User
            User object to create.

        Returns
        -------
        user: User
            The requested `User` but with the guid assigned.
        """

        # The inbound User will be missing a guid.
        # This is auto assigned on the database side and returned by the call.

        args: list[str] = [user.username, user.email, user.hashed_password, user.disabled, user.admin]
        proc_name: str = "createUser"

        rows: Optional[list[Tuple]] = self._call_proc(proc_name, args, True)
        if rows is None or len(rows) != 1:
            raise IncorrectRowCountError(f"Result count was not exactly one. Received: {rows}")
        user_guid: str = rows[0][0]

        user.guid = user_guid
        return user



""" The Rowan Tree Auth Service Definition """

import logging
import os
from pathlib import Path

from fastapi import Depends, FastAPI, Form, status
from fastapi.security import OAuth2PasswordRequestForm
from starlette.exceptions import HTTPException
from starlette.middleware.cors import CORSMiddleware

from rowantree.auth.sdk import RegisterUserRequest, Token, UserBase
from rowantree.common.sdk import demand_env_var

from ..controllers.register import RegisterController
from ..controllers.token import TokenController
from ..services.auth import AuthService
from ..services.db.dao import DBDAO
from ..services.db.utils import WrappedConnectionPool

# Setup logging
Path(demand_env_var(name="LOGS_DIR")).mkdir(parents=True, exist_ok=True)
logging.basicConfig(
    format="%(asctime)s - %(levelname)s - %(message)s",
    datefmt="%m/%d/%Y %I:%M:%S %p",
    level=logging.DEBUG,
    filemode="w",
    filename=f"{demand_env_var(name='LOGS_DIR')}/{os.uname()[1]}.therowantree.auth.service.log",
)

logging.debug("Starting server")

# Creating database connection pool, and DAO
wrapped_cnxpool: WrappedConnectionPool = WrappedConnectionPool()
dao: DBDAO = DBDAO(cnxpool=wrapped_cnxpool.cnxpool)
auth_service: AuthService = AuthService(dao=dao)


# Create controllers
token_controller: TokenController = TokenController(auth_service=auth_service)
register_controller: RegisterController = RegisterController(auth_service=auth_service)


# Create the FastAPI application
app = FastAPI()

#  Apply COR Configuration | https://fastapi.tiangolo.com/tutorial/cors/
origins = ["*"]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


# Define our handlers


@app.get("/health/plain", status_code=status.HTTP_200_OK)
def health_plain() -> bool:
    """
    Get Application Health
    [GET] /health/plain

    Returns
    -------
    health: bool
        A true/false response of server health.

    [STATUS CODE] 200: OK
    [STATUS CODE] 400: Invalid Request
    [STATUS CODE] 500: Internal Server Error
    """

    return True


@app.post("/v1/auth/token", status_code=status.HTTP_200_OK)
def token_handler(form_data: OAuth2PasswordRequestForm = Depends()) -> Token:
    """
    Get Token
    [POST] /v1/auth/token

    Returns
    -------
    token: Token

    [STATUS CODE] 200: OK
    [STATUS CODE] 400: Invalid Request
    [STATUS CODE] 401: Permission Denied
    [STATUS CODE] 500: Internal Server Error
    """
    try:
        return token_controller.execute(request=form_data)
    except HTTPException as error:
        logging.error(str(error))
        raise error from error
    except Exception as error:
        # Caught all other uncaught errors.
        logging.error(str(error))
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Internal Server Error"
        ) from error


@app.post("/v1/auth/register", status_code=status.HTTP_200_OK)
def register_handler(username: str = Form(), email: str = Form(), password: str = Form()) -> UserBase:
    """
    Register User
    [POST] /v1/auth/register

    Form Data
    ---------
    username: str
    email: str
    password: str

    Returns
    -------
    user: UserBase

    [STATUS CODE] 200: OK
    [STATUS CODE] 400: Invalid Request
    [STATUS CODE] 500: Internal Server Error
    """

    try:
        form_data: RegisterUserRequest = RegisterUserRequest(username=username, email=email, password=password)
        return register_controller.execute(request=form_data)
    except HTTPException as error:
        logging.error(str(error))
        raise error from error
    except Exception as error:
        # Caught all other uncaught errors.
        logging.error(str(error))
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Internal Server Error"
        ) from error



""" Authentication and Authorization Service Definition """

import logging
from typing import Optional

from rowantree.auth.sdk import RegisterUserRequest, User, get_password_hash, verify_password

from .abstract_service import AbstractService
from .db.incorrect_row_count_error import IncorrectRowCountError


class AuthService(AbstractService):
    """
    Authentication and Authorization Service
    """

    def authenticate_user(self, username: str, password: str) -> Optional[User]:
        """
        Authenticates a user by username/password.

        Parameters
        ----------
        username: str
            Username
        password:
            Plain text password.

        Returns
        -------
        user: Optional[User]
            Returns an instance of `UserInDB` if successfully authenticated, otherwise `None`.
        """

        try:
            user: Optional[User] = self.dao.get_user(username=username)
        except IncorrectRowCountError as error:
            logging.debug("User not found: %s", str(error))
            return None

        if not user:
            return None
        if not verify_password(password, user.hashed_password):
            return None

        return user

    # TODO: Future functionality.
    # def get_user_by_jwt(self, token: str) -> User:
    #     credentials_exception: HTTPException = HTTPException(
    #         status_code=status.HTTP_401_UNAUTHORIZED,
    #         detail="Could not validate credentials",
    #         headers={"WWW-Authenticate": "Bearer"},
    #     )
    #     user: Optional[User] = None
    #     token_claims: TokenClaims = get_claims(token=token)
    #
    #     try:
    #         user = self.dao.get_user(guid=token_claims.sub)
    #     except IncorrectRowCountError as error:
    #         logging.debug(f"User not found: {str(error)}")
    #
    #     if user is None:
    #         raise credentials_exception
    #
    #     return user

    def register_user(self, request: RegisterUserRequest) -> Optional[User]:
        """
        Creates a user record in the database.

        Parameters
        ----------
        request: RegisterUserRequest
            The user registration request.

        Returns
        -------
        user: Optional[User]
            A user if one could be created, `None` otherwise.
        """

        user: User = User(
            username=request.username,
            hashed_password=get_password_hash(request.password),
            email=request.email,
            disabled=False,
            admin=False,
        )

        try:
            return self.dao.create_user(user=user)
        except IncorrectRowCountError as error:
            logging.debug("Unable to create user, possibly already exists: %s", str(error))
            return None
